from .Objects.osu.PathType import PathType
from .PathApproximator import PathApproximator
from .Objects.Vector2 import Vector2
from .Precision import Precision

class SliderPath:
    pathType = None
    controlPoints = []
    expectedDistance = None
    isInitialised = False

    calculatedPath = []
    cumulativeLength = []
    pathApproximator = PathApproximator()

    def __init__(self, pathType, controlPoints, expectedDistance):
        self.pathType = pathType
        self.controlPoints = controlPoints
        self.expectedDistance = expectedDistance

        self.ensureInitialised()

    def ensureInitialised(self):
        if(self.isInitialised):
            return

        self.isInitialised = True
        self.controlPoints = self.controlPoints if self.controlPoints is not None else []
        self.calculatedPath = []
        self.cumulativeLength = []

        self.calculatePath()
        self.calculateCumulativeLength()

    def calculatePath(self):
        self.calculatedPath = []
        start = 0
        end = 0

        for i in range(len(self.controlPoints)):
            end += 1
            if(i == (len(self.controlPoints) - 1) or (self.controlPoints[i].x == self.controlPoints[i + 1].x and self.controlPoints[i].y == self.controlPoints[i + 1].y)):
                cpSpan = self.controlPoints[start:end]
                for t in self.calculateSubPath(cpSpan):
                    if(len(self.calculatedPath) == 0 or self.calculatedPath[len(self.calculatedPath) - 1].x != t.x or self.calculatedPath[len(self.calculatedPath) - 1].y != t.y):
                        self.calculatedPath.append(Vector2(t.x, t.y))
                
                start = end

    def calculateSubPath(self, subControlPoints):
        if(self.pathType == PathType.Linear.name):
            return self.pathApproximator.approximateLinear(subControlPoints)

        elif(self.pathType == PathType.PerfectCurve.name):
            if(len(self.controlPoints) != 3 or len(subControlPoints) != 3):
                return
            subPath = self.pathApproximator.approximateCircularArc(subControlPoints)
            if(len(subPath) == 0):
                return 
            return subPath

        elif(self.pathType == PathType.Catmull):
            return self.pathApproximator.approximateCatmull(subControlPoints)
        
        else:
            return self.pathApproximator.approximateBezier(subControlPoints)

    def calculateCumulativeLength(self):
        l = 0
        self.cumulativeLength = []
        self.cumulativeLength.append(l)

        for i in range((len(self.calculatedPath) - 1)):
            diff = self.calculatedPath[i + 1].subtract(self.calculatedPath[i])
            d = diff.length()

            if(self.expectedDistance is not None and self.expectedDistance - l < d):
                self.calculatedPath[i + 1] = self.calculatedPath[i].add(diff.scale((self.expectedDistance - l) / d))
                for j in range(i + 2, (len(self.calculatedPath) - 2 - i)):
                    self.calculatedPath.remove(self.calculatedPath[j])
                
                l = self.expectedDistance
                self.cumulativeLength.append(l)
                break

            l += d
            self.cumulativeLength.append(l)

        if(self.expectedDistance is not None and l < self.expectedDistance and len(self.calculatedPath) > 1):
            diff = self.calculatedPath[len(self.calculatedPath) - 1].subtract(self.calculatedPath[len(self.calculatedPath) - 2])
            d = diff.length()

            if(d <= 0):
                return

            self.calculatedPath[len(self.calculatedPath) - 1].add(diff.scale((self.expectedDistance - l) / d))
            self.cumulativeLength[len(self.calculatedPath) - 1] = self.expectedDistance

    def PositionAt(self, progress):
        self.ensureInitialised()
        d = self.__progressToDistance(progress)
        return self.__interpolateVertices(self.__indexOfDistance(d), d)
    
    def __progressToDistance(self, progress):
        return min(max(progress, 0), 1) * self.expectedDistance

    def __interpolateVertices(self, i, d):
        if(len(self.calculatedPath) == 0):
            return Vector2(0, 0)
        
        if(i <= 0):
            return self.calculatedPath[0]
        if(i >= len(self.calculatedPath)):
            return self.calculatedPath[len(self.calculatedPath) - 1]
        
        p0 = self.calculatedPath[i - 1]
        p1 = self.calculatedPath[i]

        d0 = self.cumulativeLength[i - 1]
        d1 = self.cumulativeLength[i]

        if(Precision.almostEqualsNumber(self, d0, d1)): # uhh what
            return p0
        
        w = (d - d0) / (d1 - d0)
        result = p0.add(p1.subtract(p0).scale(w))
        return result
    
    def __indexOfDistance(self, d):
        try:
            index = self.cumulativeLength.index(d)
        except ValueError:
            # print(ValueError)
            for cL in self.cumulativeLength:
                if(cL > d):
                    return self.cumulativeLength.index(cL)

            return len(self.cumulativeLength)

        return index


from .Objects.Vector2 import Vector2
from .Precision import Precision
from math import atan2, pi, ceil, acos, cos, sin

class PathApproximator(object):
    bezier_tolerance = 0.25
    catmull_detail = 50
    circular_arc_tolerance = 0.1

    def approximateBezier(self, controlPoints):
        output = []
        count = len(controlPoints)

        if(count == 0):
            return output
        
        subdivisionBuffer1 = []
        subdivisionBuffer2 = []
        for i in range(count):
            subdivisionBuffer1.append(Vector2(0, 0))

        for i in range(count * 2 - 1):
            subdivisionBuffer2.append(Vector2(0, 0))

        toFlatten = []
        freeBuffers = []

        deepCopy = []

        for c in controlPoints:
            deepCopy.append(Vector2(c.x, c.y))

        toFlatten.append(deepCopy)

        leftChild = subdivisionBuffer2

        while(len(toFlatten) > 0):
            parent = toFlatten.pop()
            if(self.__bezierIsFlatEnough(parent)):
                self.__bezierApproximate(parent, output, subdivisionBuffer1, subdivisionBuffer2, count)
                freeBuffers.append(parent)
                continue
        
            rightChild = []
            if(len(freeBuffers) > 0):
                rightChild = freeBuffers.pop()
            else:
                for i in range(count):
                    rightChild.append(Vector2(0, 0))

            self.__bezierSubdivide(parent, leftChild, rightChild, subdivisionBuffer1, count)

            for i in range(count):
                parent[i] = leftChild[i]

            toFlatten.append(rightChild)
            toFlatten.append(parent)
        
        output.append(controlPoints[count - 1])
        return output

    def approximateCatmull(self, controlPoints):
        result = []

        for i in range(len(controlPoints) - 1):
            v1 = controlPoints[i - 1] if i > 0 else controlPoints[i]
            v2 = controlPoints[i]
            v3 = controlPoints[i + 1] if i < (len(controlPoints) - 1) else v2.add(v2).subtract(v1)
            v4 = controlPoints[i + 2] if i < (len(controlPoints) - 2) else v3.add(v3).subtract(v2)

            for c in range(self.catmull_detail):
                result.append(self.__catmullFindPoint(v1, v2, v3, v4, c / self.catmull_detail))
                result.append(self.__catmullFindPoint(v1, v2, v3, v4, (c + 1) / self.catmull_detail))

        return result

    def approximateCircularArc(self, controlPoints):
        a = controlPoints[0]
        b = controlPoints[1]
        c = controlPoints[2]

        aSq = (b.subtract(c)).lengthSquared()
        bSq = (a.subtract(c)).lengthSquared()
        cSq = (a.subtract(b)).lengthSquared()

        if(Precision.almostEqualsNumber(aSq, 0) or Precision.almostEqualsNumber(bSq, 0) or Precision.almostEqualsNumber(cSq, 0)):
            return []
        
        s = aSq * (bSq + cSq - aSq)
        t = bSq * (aSq + cSq - bSq)
        u = cSq * (aSq + bSq - cSq)

        Sum = s + t + u

        if(Precision.almostEqualsNumber(Sum, 0)):
            return []

        centre = (a.scale(s).add(b.scale(t)).add(c.scale(u))).divide(Sum)
        dA = a.subtract(centre)
        dC = c.subtract(centre)

        r = dA.length()

        thetaStart = atan2(dA.y, dA.x)
        thetaEnd = atan2(dC.y, dC.x)

        while(thetaEnd < thetaStart):
            thetaEnd += 2 * pi

        Dir = 1
        thetaRange = thetaEnd - thetaStart

        orthoAtoC = c.subtract(a)
        orthoAtoC = Vector2(orthoAtoC.y, -1 * orthoAtoC.x)
        if(orthoAtoC.dot(b.subtract(a)) < 0):
            Dir = -1 * Dir
            thetaRange = 2 * pi - thetaRange

        amountPoints = 2 if 2 * r <= self.circular_arc_tolerance else max(2, ceil(thetaRange / (2 * acos(1 - self.circular_arc_tolerance / r))))
        
        output = []

        for i in range(amountPoints):
            fract = 1 / (amountPoints - 1)
            theta = thetaStart + Dir * fract * thetaRange
            o = Vector2(cos(theta), sin(theta)).scale(r)
            output.append(centre.add(o))

        return output

    def approximateLinear(self, controlPoints):
        return controlPoints
    
    def __bezierIsFlatEnough(self, controlPoints):
        for i in range(1, (len(controlPoints) - 1)):
            if((controlPoints[i - 1].subtract(controlPoints[i].scale(2)).add(controlPoints[i + 1])).lengthSquared() > self.bezier_tolerance * self.bezier_tolerance * 4):
                return False

        return True

    def __bezierApproximate(self, controlPoints, output, subdivisionBuffer1, subdivisionBuffer2, count):
        l = subdivisionBuffer2
        r = subdivisionBuffer1

        self.__bezierSubdivide(controlPoints, l, r, subdivisionBuffer1, count)

        for i in range(count - 1):
            l[count + i] = r[i + 1]

        output.append(controlPoints[0])
        for i in range(1, count - 1):
            index = 2 * i 
            p = (l[index - 1].add(l[index].scale(2)).add(l[index + 1])).scale(0.25)
            output.append(p)

    def __bezierSubdivide(self, controlPoints, l, r, subdivisionBuffer, count):
        midpoints = subdivisionBuffer

        for i in range(count):
            midpoints[i] = controlPoints[i]

        for i in range(count):
            l[i] = midpoints[0]
            r[count - i - 1] = midpoints[count - i - 1]

            for j in range(count - i - 1):
                midpoints[j] = (midpoints[j].add(midpoints[j + 1])).divide(2)

    def __catmullFindPoint(self, vec1, vec2, vec3, vec4, t):
        t2 = t * t
        t3 = t * t2
        result = Vector2(
            0.5 * (2 * vec2.x + (-vec1.x + vec3.x) * t + (2 * vec1.x - 5 * vec2.x + 4 * vec3.x - vec4.x) * t2 + (-vec1.x + 3 * vec2.x - 3 * vec3.x + vec4.x) * t3),
            0.5 * (2 * vec2.y + (-vec1.y + vec3.y) * t + (2 * vec1.y - 5 * vec2.y + 4 * vec3.y - vec4.y) * t2 + (-vec1.y + 3 * vec2.y - 3 * vec3.y + vec4.y) * t3)
        )

        return result



from .Objects.osu.HitObjects.DifficultyHitObject import DifficultyHitObject
from .Skills.Aim import Aim
from .Skills.Speed import Speed
from math import ceil, sqrt

class StarRatingCalculator(object):
    hitObjects = []
    section_length = 400
    difficlty_multiplier = 0.0675

    def calculate(self, hitObjects, timeRate):
        self.hitObjects = hitObjects
        aimSkill = Aim()
        speedSkill = Speed()

        sectionLength = self.section_length * timeRate

        currentSectionEnd = ceil((self.hitObjects[0].StartTime if hitObjects[0] is not None else 0) / sectionLength) * sectionLength

        for h in hitObjects:
            while(h.CurrentObject.StartTime > currentSectionEnd):
                aimSkill.saveCurrentPeak()
                aimSkill.startNewSectionFrom(currentSectionEnd)

                speedSkill.saveCurrentPeak()
                speedSkill.startNewSectionFrom(currentSectionEnd)

                currentSectionEnd += sectionLength

            aimSkill.process(h)
            speedSkill.process(h)
        
        aimSkill.saveCurrentPeak()
        speedSkill.saveCurrentPeak()

        aimRating = sqrt(aimSkill.difficultyValue()) * self.difficlty_multiplier
        speedRating = sqrt(speedSkill.difficultyValue()) * self.difficlty_multiplier
        starRating = aimRating + speedRating + abs(aimRating - speedRating) / 2
        return {
            "aim": aimRating,
            "speed": speedRating,
            "total": starRating
        }


from .Objects.osu.HitType import HitType
from .Objects.osu.PathType import PathType
from .Objects.Vector2 import Vector2
from .Objects.osu.Beatmap import Beatmap
from .Objects.osu.HitObjects.HitObject import HitObject
from .Objects.osu.HitObjects.HitCircle import HitCircle
from .Objects.osu.HitObjects.Slider import Slider
from .Objects.osu.HitObjects.Spinner import Spinner
from .SliderPath import SliderPath
from .Precision import Precision
from functools import cmp_to_key
from warnings import warn

class BeatmapParser(object):
    beatmap = Beatmap()

    def parseBeatmap(self, data, mods):
        if(not data):
            raise Exception("No beatmap found")

        self.beatmap = Beatmap()
        section = None
        lines = map(lambda line: line.strip(), data.split("\n"))
        
        for line in lines:
            # print(line)
            if(line.startswith('//') or not line):
                continue
            if(not section and 'osu file format v' in line):
                self.beatmap.Version = int(line.split('osu file format v')[1])
                continue
            
            if(line[0] == '['):
                section = line.split('[')[1].split(']')[0]
                continue
                
            if(section == 'General'):
                key, value = map(lambda v: v.strip(), line.split(':'))
                if(key == 'StackLeniency'):
                    self.beatmap.StackLeniency = float(value)
            
            elif(section == 'Difficulty'):
                key, value = map(lambda v: v.strip(), line.split(':'))
                self.beatmap.Difficulty[key] = float(value)

            elif(section == 'TimingPoints'):
                split = line.split(',')

                time = int(split[0]) + (24 if self.beatmap.Version < 5 else 0)
                beatLength = float(split[1])
                speedMultiplier = 100 / -beatLength if beatLength < 0 else 1
                timeSignature = 4
                if(len(split) >= 3):
                    timeSignature = 4 if split[2][0] == '0' else int(split[2])

                timingChange = True
                if(len(split) >= 7):
                    timingChange = split[6][0] == '1'

                if(timingChange):
                    self.beatmap.TimingPoints.append({
                        "Time": time,
                        "BeatLength": beatLength,
                        "TimeSignature": timeSignature
                    })

                self.beatmap.DifficultyTimingPoints.append({
                    "Time": time,
                    "SpeedMultiplier": speedMultiplier
                })

            elif(section == 'HitObjects'):
                split = line.split(',')
                pos = Vector2(int(split[0]), int(split[1]))
                startTime = int(split[2])
                hitType = int(split[3])

                result = None

                scale = (1 - 0.7 * (self.getCircleSize(mods) - 5) / 5) / 2
                radius = 64 * scale

                if(hitType & HitType.Normal.value[0]):
                    result = self.createCircle(pos, startTime, radius)
                
                if(hitType & HitType.Slider.value[0]):
                    pathType = None
                    length = 0
                    pointSplit = split[5].split('|')

                    points = [Vector2(0, 0)]
                    for point in pointSplit:
                        if(len(point) == 1):
                            if(point == 'C'):
                                pathType = PathType.Catmull
                            elif(point == 'B'):
                                pathType = PathType.Bezier
                            elif(point == 'L'):
                                pathType = PathType.Linear
                            elif(point == 'P'):
                                pathType = PathType.PerfectCurve
                            else:
                                pathType = PathType.Catmull
                        
                            break

                        temp = point.split(':')
                        points.append(Vector2(int(temp[0]), int(temp[1])).subtract(pos))
                    
                    def isLinear(p):
                        return Precision.almostEqualsNumber(0, (p[1].y - p[0].y) * (p[2].x - p[0].x) - (p[1].x - p[0].x) * (p[2].y - p[0].y))
                    
                    if(len(points) == 3 and pathType == PathType.PerfectCurve and isLinear(points)):
                        pathType = PathType.Linear
                    
                    repeatCount = int(split[6])
                    repeatCount = max(0, repeatCount - 1)

                    if(len(split) > 7):
                        length = float(split[7])
                    
                    result = self.createSlider(pos, points, length, pathType, repeatCount, startTime, radius)
                
                if(hitType & HitType.Spinner.value[0]):
                    endTime = int(split[5])
                    result = self.createSpinner(pos, startTime, endTime)

                self.beatmap.HitObjects.append(result)

        for h in self.beatmap.HitObjects:
            h.StackHeight = 0

        self.applyStacking(0, len(self.beatmap.HitObjects) - 1)

        scale = (1 - 0.7 * (self.getCircleSize(mods) - 5) / 5) / 2

        for hitObject in self.beatmap.HitObjects:
            hitObject.calculateStackedPosition(scale)
        
        return self.beatmap

    def createCircle(self, pos, startTime, radius):
        return HitCircle(pos, startTime, radius=radius)

    def createSlider(self, pos, points, length, pathType, repeatCount, startTime, radius):
        path = SliderPath(pathType, points, max(0, length))

        speedMultiplier = self.getSpeedMultiplier(startTime)
        beatLength = self.getBeatLength(startTime)

        return Slider(pos, startTime, path, repeatCount, speedMultiplier, beatLength, self.beatmap.Difficulty, radius)

    def createSpinner(self, pos, startTime, endTime):
        return Spinner(pos, startTime, endTime)
    
    def getSpeedMultiplier(self, startTime):
        currentTimingPoint = self.getTimingPoints(startTime, self.beatmap.DifficultyTimingPoints)
        return currentTimingPoint['SpeedMultiplier']

    def getBeatLength(self, startTime):
        currentTimingPoint = self.getTimingPoints(startTime, self.beatmap.TimingPoints)
        return currentTimingPoint['BeatLength']
    
    def getTimingPoints(self, startTime, timingPoints):
        timingPoints = sorted(timingPoints, key=cmp_to_key(lambda a, b: a['Time'] - b['Time']))
        currentTimingPoint = None
        for i in range(len(timingPoints)):
            if(timingPoints[i]['Time'] > startTime):
                currentTimingPoint = i - 1
                break
        
        if(currentTimingPoint == None):
            currentTimingPoint = len(timingPoints) - 1

        if(currentTimingPoint < 0):
            currentTimingPoint = 0
            warn(f'Warning: first timing point after current hit object ({startTime}). Defaulting to first timing point of the map.', Warning)

        return timingPoints[currentTimingPoint]
    
    def applyStacking(self, startIndex, endIndex):
        stack_distance = 3
        TimePreempt = 600

        if(self.beatmap.Difficulty['ApproachRate'] > 5):
            TimePreempt = 1200 + (450 - 1200) * (self.beatmap.Difficulty["ApproachRate"] - 5) / 5
        elif(self.beatmap.Difficulty['ApproachRate'] < 5):
            TimePreempt = 1200 - (1200 - 1800) * (5 - self.beatmap.Difficulty["ApproachRate"]) / 5
        else:
            TimePreempt = 1200

        extendedEndIndex = endIndex
        if(endIndex < (len(self.beatmap.HitObjects) - 1)):
            for i in range(endIndex, startIndex - 1, -1):
                stackBaseIndex = i
                for n in range(stackBaseIndex + 1, len(self.beatmap.HitObjects)):
                    stackBaseObject = self.beatmap.HitObjects[stackBaseIndex]
                    if(isinstance(stackBaseObject, Spinner)):
                        break

                    objectN = self.beatmap.HitObjects[n]
                    if(isinstance(objectN, Spinner)):
                        continue

                    if(isinstance(stackBaseObject, HitCircle)):
                        endTime = stackBaseObject.StartTime
                    else:
                        endTime = stackBaseObject.EndTime

                    stackThreshold = TimePreempt * self.beatmap.StackLeniency

                    if(objectN.StartTime - endTime > stackThreshold):
                        break

                    if(isinstance(stackBaseObject, Slider) and stackBaseObject.EndPosition.distance(objectN.Position) < stack_distance):
                        endPositionDistanceCheck = True
                    else:
                        endPositionDistanceCheck = False

                    if(stackBaseObject.Position.distance(objectN.Position) < stack_distance or endPositionDistanceCheck):
                        stackBaseIndex = n
                        objectN.StackHeight = 0
                
                if(stackBaseIndex > extendedEndIndex):
                    extendedEndIndex = stackBaseIndex
                    if(extendedEndIndex == (len(self.beatmap.HitObjects) - 1)):
                        break
                
        extendedStartIndex = startIndex
        for i in range(extendedEndIndex, startIndex, -1):
            n = i

            objectI = self.beatmap.HitObjects[i]
            if(objectI.StackHeight != 0 or isinstance(objectI, Spinner)):
                continue

            stackThreshold = TimePreempt * self.beatmap.StackLeniency
            if(isinstance(objectI, HitCircle)):
                while(n - 1 >= 0):
                    objectN = self.beatmap.HitObjects[n]
                    if(isinstance(objectN, Spinner)):
                        continue

                    if(isinstance(objectN, HitCircle)):
                        endTime = objectN.StartTime
                    else:
                        endTime = objectN.EndTime

                    if(objectI.StartTime - endTime > stackThreshold):
                        break

                    if(n < extendedStartIndex):
                        objectN.StackHeight = 0
                        extendedStartIndex = n

                    if(isinstance(objectN, Slider) and objectN.EndPosition.distance(objectI.Position) < stack_distance):
                        endPositionDistanceCheck = True
                    else:
                        endPositionDistanceCheck = False

                    if(endPositionDistanceCheck):
                        offset = objectI.StackHeight - objectN.StackHeight + 1
                        for j in range(n + 1, i + 1):
                            objectJ = self.beatmap.HitObjects[j]
                            if(objectN.EndPosition.distance(objectJ.Position) < stack_distance):
                                objectJ.StackHeight -= offset
                                
                        break

                    if(objectN.Position.distance(objectI.Position) < stack_distance):
                        objectN.StackHeight = objectI.StackHeight + 1
                        objectI = objectN
                    
                    n -= 1
            
            elif(isinstance(objectI, Slider)):
                while(n - 1 >= startIndex):
                    objectN = self.beatmap.HitObjects[n]
                    if(isinstance(objectN, Spinner)):
                        continue

                    if(objectI.StartTime - objectN.StartTime > stackThreshold):
                        break

                    if(isinstance(objectN, HitCircle)):
                        objectNEndPosition = objectN.Position
                    else:
                        objectNEndPosition = objectN.EndPosition

                    if(objectNEndPosition.distance(objectI.Position) < stack_distance):
                        objectN.StackHeight = objectI.StackHeight + 1
                        objectI = objectN

                    n -= 1
    
    def getCircleSize(self, mods):
        if("HR" in mods):
            return min(self.beatmap.Difficulty["CircleSize"] * 1.3, 10)
        if("EZ" in mods):
            return self.beatmap.Difficulty["CircleSize"] * 0.5
        return self.beatmap.Difficulty["CircleSize"]


from .OsuService import OsuService
from .BeatmapParser import BeatmapParser
from .DifficultyHitObjectCreator import DifficultyHitObjectCreator
from .StarRatingCalculator import StarRatingCalculator

osuService = OsuService()
beatmapParser = BeatmapParser()
difficultyHitObjectCreator = DifficultyHitObjectCreator()
starRatingCalculator = StarRatingCalculator()
Beatmap = None

def calculateStarRating(returnAllDifficultyValues = False, allCombinations = False, **kwargs):
    """Parameters:
    returnAllDifficultyValues = False
        returns total star rating value if False
        when set to True, method will also return aim and speed difficulty
        
    filepath: string
        path to .osu file (no need if map_id is set)

    map_id: integer
        BeatmapID number of a beatmap (no need if filepath is set)

    mods (optional): list of string
        Specify which mods to include during star rating calculation
        examples: 
            mods=['DT']
            mods=['EZ', 'HD', 'DT']
            mods=[]

    allCombinations = False
        when set to True, will return star rating of every possible mod combination
    """

    map_filepath = kwargs.get('filepath', None)
    map_id = kwargs.get('map_id', None)
    mods = kwargs.get('mods', None)
    allCombinations = kwargs.get('allCombinations', False)

    if(not map_filepath and not map_id):
        raise Exception('Neither BeatmapID nor beatmap filepath specified.')
    elif(map_id):
        Map = getOsuBeatmap(map_id)
    else:
        Map = getLocalOsuBeatmap(map_filepath)

    if(Map == None):
        raise Exception('No map found for specified map id. / wrong filepath')
    
    mods = parseMods(mods)
    output = {}
    if(not allCombinations):
        label = ''.join(mods) if len(mods) > 0 else "nomod"
        response = calculateNextModCombination(Map, mods, True)
        output[label] = response if returnAllDifficultyValues else response['total']
        return output
    else:
        allModCombinations = getAllModCombinations()
        for combi in allModCombinations:
            label = ''.join(combi['mods']) if len(combi['mods']) > 0 else 'nomod'
            response = calculateNextModCombination(Map, combi['mods'], combi['reParse'])
            output[label] = response if returnAllDifficultyValues else response['total']
        
        return output

def calculateNextModCombination(Map, mods, reParse):
    if(reParse):
        Beatmap = beatmapParser.parseBeatmap(Map, mods)

    timeRate = getTimeRate(mods)
    difficultyHitObjects = difficultyHitObjectCreator.convertToDifficultyHitObjects(Beatmap.HitObjects, timeRate)
    return starRatingCalculator.calculate(difficultyHitObjects, timeRate)

def getOsuBeatmap(map_id):
    return osuService.getOsuBeatmap(map_id)

def parseMods(mods):
    if(mods == None):
        return []
    return mods

def getTimeRate(mods):
    if("DT" in mods):
        return 1.5
    if("HT" in mods):
        return 0.75
    return 1

def getAllModCombinations():
    return [
        { 'mods': [], 'reParse': True },
        { 'mods': ['DT'], 'reParse': False },
        { 'mods': ['HT'], 'reParse': False },
        { 'mods': ['HR'], 'reParse': True },
        { 'mods': ['HR', 'DT'], 'reParse': False },
        { 'mods': ['HR', 'HT'], 'reParse': False },
        { 'mods': ['EZ'], 'reParse': True },
        { 'mods': ['EZ', 'DT'], 'reParse': False },
        { 'mods': ['EZ', 'HT'], 'reParse': False },
    ]

def getLocalOsuBeatmap(filePath):
    return open(filePath, 'r').read()


from .Objects.Vector2 import Vector2
from abc import ABC

class Precision(ABC):
    FLOAT_EPSILON = 1e-3

    def almostEqualsNumber(self, value1, value2, acceptableDifference = FLOAT_EPSILON):
        return abs(value1 - value2) <= acceptableDifference

    def almostEqualsVector(self, vec1, vec2, acceptableDifference = FLOAT_EPSILON):
        return self.almostEqualsNumber(vec1.x, vec2.x, acceptableDifference) and self.almostEqualsNumber(vec1.y, vec2.y, acceptableDifference)



from requests import get

# just downloads the .osu file from osu servers
class OsuService(object):
    def getOsuBeatmap(self, map_id):
        url = "https://osu.ppy.sh/osu/" + str(map_id)
        req = get(url)
        if(req):
            return req.text # check that later
        else:
            raise Exception("OsuService failed map_id GET request")


from .Objects.osu.HitObjects.HitObject import HitObject
from .Objects.osu.HitObjects.DifficultyHitObject import DifficultyHitObject
from .Objects.osu.HitObjects.Slider import Slider
from .Objects.osu.HitObjects.Spinner import Spinner
from .Objects.Vector2 import Vector2
from math import atan2

class DifficultyHitObjectCreator(object):
    difficultyHitObjects = []

    lastLastObject = None
    lastObject = None
    currentObject = None
    timeRate = None

    normalized_radius = 52
    TravelDistance = 0
    JumpDistance = 0
    Angle = 0
    DeltaTime = 0
    StrainTime = 0

    def convertToDifficultyHitObjects(self, hitObjects, timeRate):
        self.difficultyHitObjects = []

        for i in range(1, len(hitObjects)):
            lastLast = hitObjects[i - 2] if i > 1 else None
            last = hitObjects[i - 1]
            current = hitObjects[i]

            difficultyHitObject = self.createDifficultyHitObject(lastLast, last, current, timeRate)
            self.difficultyHitObjects.append(difficultyHitObject)

        return self.difficultyHitObjects

    def createDifficultyHitObject(self, lastLast, last, current, timeRate):
        self.lastLastObject = lastLast
        self.lastObject = last
        self.currentObject = current
        self.timeRate = timeRate

        self.setDistances()
        self.setTimingValues()

        return DifficultyHitObject(self.currentObject, self.lastObject, self.lastLastObject, self.TravelDistance, self.JumpDistance, self.Angle, self.DeltaTime, self.StrainTime)

    def setDistances(self):
        self.TravelDistance = 0
        self.JumpDistance = 0
        self.Angle = 0
        self.DeltaTime = 0
        self.StrainTime = 0

        scalingFactor = self.normalized_radius / self.currentObject.Radius  
        if(self.currentObject.Radius < 30):
            smallCircleBonus = min(30 - self.currentObject.Radius, 5) / 50
            scalingFactor *= 1 + smallCircleBonus
        
        if(isinstance(self.lastObject, Slider)):
            lastSlider = self.lastObject
            self.computeSliderCursorPosition(lastSlider)
            self.TravelDistance = lastSlider.LazyTravelDistance * scalingFactor

        lastCursorPosition = self.getEndCursorPosition(self.lastObject)

        if(not isinstance(self.currentObject, Spinner)):
            self.JumpDistance = self.currentObject.StackedPosition.scale(scalingFactor).subtract(lastCursorPosition.scale(scalingFactor)).length()
        
        if(self.lastLastObject is not None):
            lastLastCursorPosition = self.getEndCursorPosition(self.lastLastObject)

            v1 = lastLastCursorPosition.subtract(self.lastObject.StackedPosition)
            v2 = self.currentObject.StackedPosition.subtract(lastCursorPosition)

            dot = v1.dot(v2)
            det = v1.x * v2.y - v1.y * v2.x

            self.Angle = abs(atan2(det, dot))

    def setTimingValues(self):
        self.DeltaTime = (self.currentObject.StartTime - self.lastObject.StartTime) / self.timeRate
        self.StrainTime = max(50, self.DeltaTime)

    def computeSliderCursorPosition(self, slider):
        if(slider.LazyEndPosition is not None):
            return
        slider.LazyEndPosition = slider.StackedPosition
        slider.LazyTravelDistance = 0

        approxFollowCircleRadius = slider.Radius * 3
        def computeVertex(t):
            progress = (t - slider.StartTime) / slider.SpanDuration
            if(progress % 2 >= 1):
                progress = 1 - progress % 1
            else:
                progress = progress % 1
            
            diff = slider.StackedPosition.add(slider.Path.PositionAt(progress)).subtract(slider.LazyEndPosition)
            dist = diff.length()

            if(dist > approxFollowCircleRadius):
                diff.normalize()
                dist -= approxFollowCircleRadius
                slider.LazyEndPosition = slider.LazyEndPosition.add(diff.scale(dist))
                slider.LazyTravelDistance = dist if slider.LazyTravelDistance is None else slider.LazyTravelDistance + dist

        def mapFunc(t):
            return t.StartTime
        scoringTimes = map(mapFunc, slider.NestedHitObjects[1:len(slider.NestedHitObjects)])

        for time in scoringTimes:
            computeVertex(time)

    def getEndCursorPosition(self, obj):
        pos = obj.StackedPosition

        if(isinstance(obj, Slider)):
            self.computeSliderCursorPosition(obj)
            pos = obj.LazyEndPosition if obj.LazyEndPosition is not None else pos

        return pos


from setuptools import setup, find_packages

readme = open('README.md', 'r').read()
req = list(map(lambda v: v.rstrip('\n'), open("requirements.txt", 'r').readlines()))

setup(
    name='osu_sr_calculator',
    author='Piggey',
    version='0.2',
    description="MrHelix's osu! star rating calculator rewritten for your Python needs",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/Piggey/osu_sr_calculator",
    install_reqs=req,
    keywords='osu!, osu, star rating, calculator, sr',
    packages=find_packages(include=['osu_sr_calculator']),
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Natural Language :: English',
        "Intended Audience :: Developers",
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3'
    ],

    python_requires='>=3.7'
)